from requests import Request, Session

from twilio.http import HttpClient, get_cert_file
from twilio.http.response import Response


class TwilioHttpClient(HttpClient):
    """
    General purpose HTTP Client for interacting with the Twilio API
    """
    def __init__(self, connection_pool=True):
        if connection_pool:
            self.session = Session()
            self.session.verify = get_cert_file()
        else: 
            self.session = None
            
    def request(self, method, url, params=None, data=None, headers=None, auth=None, timeout=None,
                allow_redirects=False):
        """
        Make an HTTP Request with parameters provided.

        :param str method: The HTTP method to use
        :param str url: The URL to request
        :param dict params: Query parameters to append to the URL
        :param dict data: Parameters to go in the body of the HTTP request
        :param dict headers: HTTP Headers to send with the request
        :param tuple auth: Basic Auth arguments
        :param float timeout: Socket/Read timeout for the request
        :param boolean allow_redirects: Whether or not to allow redirects
        See the requests documentation for explanation of all these parameters

        :return: An http response
        :rtype: A :class:`Response <twilio.rest.http.response.Response>` object
        """
        session = self.session
        if session is None:
            session = Session()
            session.verify = get_cert_file()

        request = Request(method.upper(), url, params=params, data=data, headers=headers, auth=auth)

        prepped_request = session.prepare_request(request)
        response = session.send(
            prepped_request,
            allow_redirects=allow_redirects,
            timeout=timeout,
        )

        return Response(int(response.status_code), response.text)



# coding=utf-8
"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page


class ConnectAppList(ListResource):
    """  """

    def __init__(self, version, account_sid):
        """
        Initialize the ConnectAppList

        :param Version version: Version that contains the resource
        :param account_sid: The unique sid that identifies this account

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppList
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppList
        """
        super(ConnectAppList, self).__init__(version)

        # Path Solution
        self._solution = {
            'account_sid': account_sid,
        }
        self._uri = '/Accounts/{account_sid}/ConnectApps.json'.format(**self._solution)

    def stream(self, limit=None, page_size=None):
        """
        Streams ConnectAppInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.api.v2010.account.connect_app.ConnectAppInstance]
        """
        limits = self._version.read_limits(limit, page_size)

        page = self.page(
            page_size=limits['page_size'],
        )

        return self._version.stream(page, limits['limit'], limits['page_limit'])

    def list(self, limit=None, page_size=None):
        """
        Lists ConnectAppInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.api.v2010.account.connect_app.ConnectAppInstance]
        """
        return list(self.stream(
            limit=limit,
            page_size=page_size,
        ))

    def page(self, page_token=values.unset, page_number=values.unset,
             page_size=values.unset):
        """
        Retrieve a single page of ConnectAppInstance records from the API.
        Request is executed immediately

        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppPage
        """
        params = values.of({
            'PageToken': page_token,
            'Page': page_number,
            'PageSize': page_size,
        })

        response = self._version.page(
            'GET',
            self._uri,
            params=params,
        )

        return ConnectAppPage(self._version, response, self._solution)

    def get_page(self, target_url):
        """
        Retrieve a specific page of ConnectAppInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppPage
        """
        response = self._version.domain.twilio.request(
            'GET',
            target_url,
        )

        return ConnectAppPage(self._version, response, self._solution)

    def get(self, sid):
        """
        Constructs a ConnectAppContext

        :param sid: Fetch by unique connect-app Sid

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        """
        return ConnectAppContext(
            self._version,
            account_sid=self._solution['account_sid'],
            sid=sid,
        )

    def __call__(self, sid):
        """
        Constructs a ConnectAppContext

        :param sid: Fetch by unique connect-app Sid

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        """
        return ConnectAppContext(
            self._version,
            account_sid=self._solution['account_sid'],
            sid=sid,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Api.V2010.ConnectAppList>'


class ConnectAppPage(Page):
    """  """

    def __init__(self, version, response, solution):
        """
        Initialize the ConnectAppPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API
        :param account_sid: The unique sid that identifies this account

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppPage
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppPage
        """
        super(ConnectAppPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of ConnectAppInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        return ConnectAppInstance(
            self._version,
            payload,
            account_sid=self._solution['account_sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Api.V2010.ConnectAppPage>'


class ConnectAppContext(InstanceContext):
    """  """

    def __init__(self, version, account_sid, sid):
        """
        Initialize the ConnectAppContext

        :param Version version: Version that contains the resource
        :param account_sid: The account_sid
        :param sid: Fetch by unique connect-app Sid

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        """
        super(ConnectAppContext, self).__init__(version)

        # Path Solution
        self._solution = {
            'account_sid': account_sid,
            'sid': sid,
        }
        self._uri = '/Accounts/{account_sid}/ConnectApps/{sid}.json'.format(**self._solution)

    def fetch(self):
        """
        Fetch a ConnectAppInstance

        :returns: Fetched ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        params = values.of({})

        payload = self._version.fetch(
            'GET',
            self._uri,
            params=params,
        )

        return ConnectAppInstance(
            self._version,
            payload,
            account_sid=self._solution['account_sid'],
            sid=self._solution['sid'],
        )

    def update(self, authorize_redirect_url=values.unset, company_name=values.unset,
               deauthorize_callback_method=values.unset,
               deauthorize_callback_url=values.unset, description=values.unset,
               friendly_name=values.unset, homepage_url=values.unset,
               permissions=values.unset):
        """
        Update the ConnectAppInstance

        :param unicode authorize_redirect_url: URIL Twilio sends requests when users authorize
        :param unicode company_name: The company name set for this Connect App.
        :param unicode deauthorize_callback_method: HTTP method Twilio WIll use making requests to the url
        :param unicode deauthorize_callback_url: URL Twilio will send a request when a user de-authorizes this app
        :param unicode description: A more detailed human readable description
        :param unicode friendly_name: A human readable name for the Connect App.
        :param unicode homepage_url: The URL users can obtain more information
        :param ConnectAppInstance.Permission permissions: The set of permissions that your ConnectApp requests.

        :returns: Updated ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        data = values.of({
            'AuthorizeRedirectUrl': authorize_redirect_url,
            'CompanyName': company_name,
            'DeauthorizeCallbackMethod': deauthorize_callback_method,
            'DeauthorizeCallbackUrl': deauthorize_callback_url,
            'Description': description,
            'FriendlyName': friendly_name,
            'HomepageUrl': homepage_url,
            'Permissions': permissions,
        })

        payload = self._version.update(
            'POST',
            self._uri,
            data=data,
        )

        return ConnectAppInstance(
            self._version,
            payload,
            account_sid=self._solution['account_sid'],
            sid=self._solution['sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Api.V2010.ConnectAppContext {}>'.format(context)


class ConnectAppInstance(InstanceResource):
    """  """

    class Permission(object):
        GET_ALL = "get-all"
        POST_ALL = "post-all"

    def __init__(self, version, payload, account_sid, sid=None):
        """
        Initialize the ConnectAppInstance

        :returns: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        super(ConnectAppInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'account_sid': payload['account_sid'],
            'authorize_redirect_url': payload['authorize_redirect_url'],
            'company_name': payload['company_name'],
            'deauthorize_callback_method': payload['deauthorize_callback_method'],
            'deauthorize_callback_url': payload['deauthorize_callback_url'],
            'description': payload['description'],
            'friendly_name': payload['friendly_name'],
            'homepage_url': payload['homepage_url'],
            'permissions': payload['permissions'],
            'sid': payload['sid'],
            'uri': payload['uri'],
        }

        # Context
        self._context = None
        self._solution = {
            'account_sid': account_sid,
            'sid': sid or self._properties['sid'],
        }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: ConnectAppContext for this ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppContext
        """
        if self._context is None:
            self._context = ConnectAppContext(
                self._version,
                account_sid=self._solution['account_sid'],
                sid=self._solution['sid'],
            )
        return self._context

    @property
    def account_sid(self):
        """
        :returns: The unique sid that identifies this account
        :rtype: unicode
        """
        return self._properties['account_sid']

    @property
    def authorize_redirect_url(self):
        """
        :returns: URIL Twilio sends requests when users authorize
        :rtype: unicode
        """
        return self._properties['authorize_redirect_url']

    @property
    def company_name(self):
        """
        :returns: The company name set for this Connect App.
        :rtype: unicode
        """
        return self._properties['company_name']

    @property
    def deauthorize_callback_method(self):
        """
        :returns: HTTP method Twilio WIll use making requests to the url
        :rtype: unicode
        """
        return self._properties['deauthorize_callback_method']

    @property
    def deauthorize_callback_url(self):
        """
        :returns: URL Twilio will send a request when a user de-authorizes this app
        :rtype: unicode
        """
        return self._properties['deauthorize_callback_url']

    @property
    def description(self):
        """
        :returns: A more detailed human readable description
        :rtype: unicode
        """
        return self._properties['description']

    @property
    def friendly_name(self):
        """
        :returns: A human readable name for the Connect App.
        :rtype: unicode
        """
        return self._properties['friendly_name']

    @property
    def homepage_url(self):
        """
        :returns: The URL users can obtain more information
        :rtype: unicode
        """
        return self._properties['homepage_url']

    @property
    def permissions(self):
        """
        :returns: The set of permissions that your ConnectApp requests.
        :rtype: ConnectAppInstance.Permission
        """
        return self._properties['permissions']

    @property
    def sid(self):
        """
        :returns: A string that uniquely identifies this connect-apps
        :rtype: unicode
        """
        return self._properties['sid']

    @property
    def uri(self):
        """
        :returns: The URI for this resource
        :rtype: unicode
        """
        return self._properties['uri']

    def fetch(self):
        """
        Fetch a ConnectAppInstance

        :returns: Fetched ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        return self._proxy.fetch()

    def update(self, authorize_redirect_url=values.unset, company_name=values.unset,
               deauthorize_callback_method=values.unset,
               deauthorize_callback_url=values.unset, description=values.unset,
               friendly_name=values.unset, homepage_url=values.unset,
               permissions=values.unset):
        """
        Update the ConnectAppInstance

        :param unicode authorize_redirect_url: URIL Twilio sends requests when users authorize
        :param unicode company_name: The company name set for this Connect App.
        :param unicode deauthorize_callback_method: HTTP method Twilio WIll use making requests to the url
        :param unicode deauthorize_callback_url: URL Twilio will send a request when a user de-authorizes this app
        :param unicode description: A more detailed human readable description
        :param unicode friendly_name: A human readable name for the Connect App.
        :param unicode homepage_url: The URL users can obtain more information
        :param ConnectAppInstance.Permission permissions: The set of permissions that your ConnectApp requests.

        :returns: Updated ConnectAppInstance
        :rtype: twilio.rest.api.v2010.account.connect_app.ConnectAppInstance
        """
        return self._proxy.update(
            authorize_redirect_url=authorize_redirect_url,
            company_name=company_name,
            deauthorize_callback_method=deauthorize_callback_method,
            deauthorize_callback_url=deauthorize_callback_url,
            description=description,
            friendly_name=friendly_name,
            homepage_url=homepage_url,
            permissions=permissions,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Api.V2010.ConnectAppInstance {}>'.format(context)



# coding=utf-8
"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import serialize
from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page


class TaskQueueStatisticsList(ListResource):
    """  """

    def __init__(self, version, workspace_sid, task_queue_sid):
        """
        Initialize the TaskQueueStatisticsList

        :param Version version: Version that contains the resource
        :param workspace_sid: The workspace_sid
        :param task_queue_sid: The task_queue_sid

        :returns: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsList
        :rtype: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsList
        """
        super(TaskQueueStatisticsList, self).__init__(version)

        # Path Solution
        self._solution = {
            'workspace_sid': workspace_sid,
            'task_queue_sid': task_queue_sid,
        }

    def get(self):
        """
        Constructs a TaskQueueStatisticsContext

        :returns: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsContext
        :rtype: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsContext
        """
        return TaskQueueStatisticsContext(
            self._version,
            workspace_sid=self._solution['workspace_sid'],
            task_queue_sid=self._solution['task_queue_sid'],
        )

    def __call__(self):
        """
        Constructs a TaskQueueStatisticsContext

        :returns: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsContext
        :rtype: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsContext
        """
        return TaskQueueStatisticsContext(
            self._version,
            workspace_sid=self._solution['workspace_sid'],
            task_queue_sid=self._solution['task_queue_sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Taskrouter.V1.TaskQueueStatisticsList>'


class TaskQueueStatisticsPage(Page):
    """  """

    def __init__(self, version, response, solution):
        """
        Initialize the TaskQueueStatisticsPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API
        :param workspace_sid: The workspace_sid
        :param task_queue_sid: The task_queue_sid

        :returns: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsPage
        :rtype: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsPage
        """
        super(TaskQueueStatisticsPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of TaskQueueStatisticsInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsInstance
        """
        return TaskQueueStatisticsInstance(
            self._version,
            payload,
            workspace_sid=self._solution['workspace_sid'],
            task_queue_sid=self._solution['task_queue_sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Taskrouter.V1.TaskQueueStatisticsPage>'


class TaskQueueStatisticsContext(InstanceContext):
    """  """

    def __init__(self, version, workspace_sid, task_queue_sid):
        """
        Initialize the TaskQueueStatisticsContext

        :param Version version: Version that contains the resource
        :param workspace_sid: The workspace_sid
        :param task_queue_sid: The task_queue_sid

        :returns: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsContext
        :rtype: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsContext
        """
        super(TaskQueueStatisticsContext, self).__init__(version)

        # Path Solution
        self._solution = {
            'workspace_sid': workspace_sid,
            'task_queue_sid': task_queue_sid,
        }
        self._uri = '/Workspaces/{workspace_sid}/TaskQueues/{task_queue_sid}/Statistics'.format(**self._solution)

    def fetch(self, end_date=values.unset, minutes=values.unset,
              start_date=values.unset):
        """
        Fetch a TaskQueueStatisticsInstance

        :param datetime end_date: The end_date
        :param unicode minutes: The minutes
        :param datetime start_date: The start_date

        :returns: Fetched TaskQueueStatisticsInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsInstance
        """
        params = values.of({
            'EndDate': serialize.iso8601_datetime(end_date),
            'Minutes': minutes,
            'StartDate': serialize.iso8601_datetime(start_date),
        })

        payload = self._version.fetch(
            'GET',
            self._uri,
            params=params,
        )

        return TaskQueueStatisticsInstance(
            self._version,
            payload,
            workspace_sid=self._solution['workspace_sid'],
            task_queue_sid=self._solution['task_queue_sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Taskrouter.V1.TaskQueueStatisticsContext {}>'.format(context)


class TaskQueueStatisticsInstance(InstanceResource):
    """  """

    def __init__(self, version, payload, workspace_sid, task_queue_sid):
        """
        Initialize the TaskQueueStatisticsInstance

        :returns: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsInstance
        """
        super(TaskQueueStatisticsInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'account_sid': payload['account_sid'],
            'cumulative': payload['cumulative'],
            'realtime': payload['realtime'],
            'task_queue_sid': payload['task_queue_sid'],
            'workspace_sid': payload['workspace_sid'],
            'url': payload['url'],
        }

        # Context
        self._context = None
        self._solution = {
            'workspace_sid': workspace_sid,
            'task_queue_sid': task_queue_sid,
        }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: TaskQueueStatisticsContext for this TaskQueueStatisticsInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsContext
        """
        if self._context is None:
            self._context = TaskQueueStatisticsContext(
                self._version,
                workspace_sid=self._solution['workspace_sid'],
                task_queue_sid=self._solution['task_queue_sid'],
            )
        return self._context

    @property
    def account_sid(self):
        """
        :returns: The account_sid
        :rtype: unicode
        """
        return self._properties['account_sid']

    @property
    def cumulative(self):
        """
        :returns: The cumulative
        :rtype: dict
        """
        return self._properties['cumulative']

    @property
    def realtime(self):
        """
        :returns: The realtime
        :rtype: dict
        """
        return self._properties['realtime']

    @property
    def task_queue_sid(self):
        """
        :returns: The task_queue_sid
        :rtype: unicode
        """
        return self._properties['task_queue_sid']

    @property
    def workspace_sid(self):
        """
        :returns: The workspace_sid
        :rtype: unicode
        """
        return self._properties['workspace_sid']

    @property
    def url(self):
        """
        :returns: The url
        :rtype: unicode
        """
        return self._properties['url']

    def fetch(self, end_date=values.unset, minutes=values.unset,
              start_date=values.unset):
        """
        Fetch a TaskQueueStatisticsInstance

        :param datetime end_date: The end_date
        :param unicode minutes: The minutes
        :param datetime start_date: The start_date

        :returns: Fetched TaskQueueStatisticsInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.task_queue.task_queue_statistics.TaskQueueStatisticsInstance
        """
        return self._proxy.fetch(
            end_date=end_date,
            minutes=minutes,
            start_date=start_date,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Taskrouter.V1.TaskQueueStatisticsInstance {}>'.format(context)



# coding=utf-8
"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page


class SegmentMembershipList(ListResource):
    """ PLEASE NOTE that this class contains beta products that are subject to
    change. Use them with caution. """

    def __init__(self, version, service_sid, identity):
        """
        Initialize the SegmentMembershipList

        :param Version version: Version that contains the resource
        :param service_sid: The service_sid
        :param identity: The identity

        :returns: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipList
        :rtype: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipList
        """
        super(SegmentMembershipList, self).__init__(version)

        # Path Solution
        self._solution = {
            'service_sid': service_sid,
            'identity': identity,
        }
        self._uri = '/Services/{service_sid}/Users/{identity}/SegmentMemberships'.format(**self._solution)

    def create(self, segment):
        """
        Create a new SegmentMembershipInstance

        :param unicode segment: The segment

        :returns: Newly created SegmentMembershipInstance
        :rtype: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipInstance
        """
        data = values.of({
            'Segment': segment,
        })

        payload = self._version.create(
            'POST',
            self._uri,
            data=data,
        )

        return SegmentMembershipInstance(
            self._version,
            payload,
            service_sid=self._solution['service_sid'],
            identity=self._solution['identity'],
        )

    def get(self, segment):
        """
        Constructs a SegmentMembershipContext

        :param segment: The segment

        :returns: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipContext
        :rtype: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipContext
        """
        return SegmentMembershipContext(
            self._version,
            service_sid=self._solution['service_sid'],
            identity=self._solution['identity'],
            segment=segment,
        )

    def __call__(self, segment):
        """
        Constructs a SegmentMembershipContext

        :param segment: The segment

        :returns: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipContext
        :rtype: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipContext
        """
        return SegmentMembershipContext(
            self._version,
            service_sid=self._solution['service_sid'],
            identity=self._solution['identity'],
            segment=segment,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Notify.V1.SegmentMembershipList>'


class SegmentMembershipPage(Page):
    """ PLEASE NOTE that this class contains beta products that are subject to
    change. Use them with caution. """

    def __init__(self, version, response, solution):
        """
        Initialize the SegmentMembershipPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API
        :param service_sid: The service_sid
        :param identity: The identity

        :returns: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipPage
        :rtype: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipPage
        """
        super(SegmentMembershipPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of SegmentMembershipInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipInstance
        :rtype: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipInstance
        """
        return SegmentMembershipInstance(
            self._version,
            payload,
            service_sid=self._solution['service_sid'],
            identity=self._solution['identity'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Notify.V1.SegmentMembershipPage>'


class SegmentMembershipContext(InstanceContext):
    """ PLEASE NOTE that this class contains beta products that are subject to
    change. Use them with caution. """

    def __init__(self, version, service_sid, identity, segment):
        """
        Initialize the SegmentMembershipContext

        :param Version version: Version that contains the resource
        :param service_sid: The service_sid
        :param identity: The identity
        :param segment: The segment

        :returns: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipContext
        :rtype: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipContext
        """
        super(SegmentMembershipContext, self).__init__(version)

        # Path Solution
        self._solution = {
            'service_sid': service_sid,
            'identity': identity,
            'segment': segment,
        }
        self._uri = '/Services/{service_sid}/Users/{identity}/SegmentMemberships/{segment}'.format(**self._solution)

    def delete(self):
        """
        Deletes the SegmentMembershipInstance

        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._version.delete('delete', self._uri)

    def fetch(self):
        """
        Fetch a SegmentMembershipInstance

        :returns: Fetched SegmentMembershipInstance
        :rtype: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipInstance
        """
        params = values.of({})

        payload = self._version.fetch(
            'GET',
            self._uri,
            params=params,
        )

        return SegmentMembershipInstance(
            self._version,
            payload,
            service_sid=self._solution['service_sid'],
            identity=self._solution['identity'],
            segment=self._solution['segment'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Notify.V1.SegmentMembershipContext {}>'.format(context)


class SegmentMembershipInstance(InstanceResource):
    """ PLEASE NOTE that this class contains beta products that are subject to
    change. Use them with caution. """

    def __init__(self, version, payload, service_sid, identity, segment=None):
        """
        Initialize the SegmentMembershipInstance

        :returns: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipInstance
        :rtype: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipInstance
        """
        super(SegmentMembershipInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'account_sid': payload['account_sid'],
            'service_sid': payload['service_sid'],
            'identity': payload['identity'],
            'segment': payload['segment'],
            'url': payload['url'],
        }

        # Context
        self._context = None
        self._solution = {
            'service_sid': service_sid,
            'identity': identity,
            'segment': segment or self._properties['segment'],
        }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: SegmentMembershipContext for this SegmentMembershipInstance
        :rtype: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipContext
        """
        if self._context is None:
            self._context = SegmentMembershipContext(
                self._version,
                service_sid=self._solution['service_sid'],
                identity=self._solution['identity'],
                segment=self._solution['segment'],
            )
        return self._context

    @property
    def account_sid(self):
        """
        :returns: The account_sid
        :rtype: unicode
        """
        return self._properties['account_sid']

    @property
    def service_sid(self):
        """
        :returns: The service_sid
        :rtype: unicode
        """
        return self._properties['service_sid']

    @property
    def identity(self):
        """
        :returns: The identity
        :rtype: unicode
        """
        return self._properties['identity']

    @property
    def segment(self):
        """
        :returns: The segment
        :rtype: unicode
        """
        return self._properties['segment']

    @property
    def url(self):
        """
        :returns: The url
        :rtype: unicode
        """
        return self._properties['url']

    def delete(self):
        """
        Deletes the SegmentMembershipInstance

        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._proxy.delete()

    def fetch(self):
        """
        Fetch a SegmentMembershipInstance

        :returns: Fetched SegmentMembershipInstance
        :rtype: twilio.rest.notify.v1.service.user.segment_memberships.SegmentMembershipInstance
        """
        return self._proxy.fetch()

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Notify.V1.SegmentMembershipInstance {}>'.format(context)



import os

from twilio.base.exceptions import TwilioException


def get_cert_file():
    """ Get the cert file location or bail """
    # XXX - this currently fails test coverage because we don't actually go
    # over the network anywhere. Might be good to have a test that stands up a
    # local server and authenticates against it.
    try:
        # Apparently __file__ is not available in all places so wrapping this
        # in a try/catch
        current_path = os.path.realpath(__file__)
        ca_cert_path = os.path.join(current_path, '..', '..', 'conf', 'cacert.pem')
        return os.path.abspath(ca_cert_path)
    except Exception:
        # None means use the default system file
        return None


class HttpClient(object):
    """
    An abstract class representing an HTTP client.
    """
    def request(self, method, url, params=None, data=None, headers=None, auth=None,
                timeout=None, allow_redirects=False):
        """
        Make an HTTP request.
        """
        raise TwilioException('HttpClient is an abstract class')



import json
from six import string_types

from twilio.twiml import TwiML, format_language


class VoiceResponse(TwiML):
    """
    Voice TwiML Response
    """
    def __init__(self):
        """
        Create a new <Response>
        """
        super(VoiceResponse, self).__init__()
        self.name = 'Response'

    def dial(self,
             number,
             action=None,
             method=None,
             timeout=None,
             hangup_on_star=None,
             time_limit=None,
             caller_id=None,
             record=None,
             trim=None,
             recording_status_callback=None,
             recording_status_callback_method=None,
             **kwargs):
        """
        Create a <Dial> element

        :param number: phone number to dial
        :param action: action URL
        :param method: action HTTP method
        :param timeout: time to wait for answer
        :param hangup_on_star: hangup call on * press
        :param time_limit: max time length
        :param caller_id: caller ID to display
        :param record: record the call
        :param trim: trim the recording
        :param recording_status_callback: status callback URL
        :param recording_status_callback_method: status callback URL method
        :param kwargs: additional attributes
        :return: <Response> element
        """
        return self.nest(Dial(
            number=number,
            action=action,
            method=method,
            timeout=timeout,
            hangup_on_star=hangup_on_star,
            time_limit=time_limit,
            caller_id=caller_id,
            record=record,
            trim=trim,
            recording_status_callback=recording_status_callback,
            recording_status_callback_method=recording_status_callback_method,
            **kwargs
        ))

    def echo(self,
             **kwargs):
        """
        Add a new <Echo> element

        :return: <Response> element
        """
        return self.nest(Echo(
            **kwargs
        ))

    def enqueue(self,
                name,
                action=None,
                method=None,
                wait_url=None,
                wait_url_method=None,
                workflow_sid=None,
                **kwargs):
        """
        Add a new <Enqueue> element

        :param name: friendly name
        :param action: action URL
        :param method: action URL method
        :param wait_url: wait URL
        :param wait_url_method: wait URL method
        :param workflow_sid: TaskRouter workflow SID
        :param kwargs: additional attributes
        :return: <Response> element
        """
        return self.nest(Enqueue(
            name,
            action=action,
            method=method,
            wait_url=wait_url,
            wait_url_method=wait_url_method,
            workflow_sid=workflow_sid,
            **kwargs
        ))

    def gather(self,
               action=None,
               method=None,
               timeout=None,
               finish_on_key=None,
               num_digits=None,
               partial_result_callback=None,
               partial_result_callback_method=None,
               language=None,
               hints=None,
               barge_in=None,
               acknowledge_sound_url=None,
               input=None,
               **kwargs):
        """
        Add a new <Gather> element

        :param action: action URL
        :param method: action URL method
        :param timeout: time to wait while gathering input
        :param finish_on_key: finish on key press
        :param num_digits: digits to collect
        :param partial_result_callback: callback url
        :param partial_result_callback_method: callback method
        :param language: locale string
        :param hints: speech recognition hints
        :param barge_in: stop playing media upon speech
        :param acknowledge_sound_url: url to hit when sound starts
        :param input: type Twilio should accept "dtfm", "speech", "dtfm speech"
        :param kwargs: additional attributes
        :return: <Response> element
        """
        return self.nest(Gather(
            action=action,
            method=method,
            timeout=timeout,
            finish_on_key=finish_on_key,
            num_digits=num_digits,
            partial_result_callback=partial_result_callback,
            partial_result_callback_method=partial_result_callback_method,
            language=format_language(language),
            hints=hints,
            barge_in=barge_in,
            acknowledge_sound_url=acknowledge_sound_url,
            input=input,
            **kwargs
        ))

    def hangup(self):
        """
        Add a new <Hangup> element

        :return: <Response> element
        """
        return self.nest(Hangup())

    def leave(self):
        """
        Add a new <Leave> element

        :return: <Response> element
        """
        return self.nest(Leave())

    def pause(self, length=None):
        """
        Add a new <Pause> element

        :param length: time in seconds to pause
        :return: <Response> element
        """
        return self.nest(Pause(length=length))

    def play(self,
             url=None,
             loop=None,
             digits=None,
             **kwargs):
        """
        Add a new <Play> element

        :param url: url to play
        :param loop: times to loop
        :param digits: play DTMF tones during a call
        :param kwargs: additional attributes
        :return: <Response> element
        """
        return self.nest(Play(
            url=url,
            loop=loop,
            digits=digits,
            **kwargs
        ))

    def record(self,
               action=None,
               method=None,
               timeout=None,
               finish_on_key=None,
               max_length=None,
               play_beep=None,
               trim=None,
               recording_status_callback=None,
               recording_status_callback_method=None,
               transcribe=None,
               transcribe_callback=None,
               **kwargs):
        """
        Add a new <Record> element

        :param action: action URL
        :param method: action URL method
        :param timeout: timeout for recording
        :param finish_on_key: finish recording on key
        :param max_length: max length to record
        :param play_beep: play beep
        :param trim: trim the recording
        :param recording_status_callback: status callback for the recordings
        :param recording_status_callback_method: status callback method
        :param transcribe: transcribe the recording
        :param transcribe_callback: transcribe callback URL
        :param kwargs: additional attributes
        :return: <Response> element
        """
        return self.nest(Record(
            action=action,
            method=method,
            timeout=timeout,
            finish_on_key=finish_on_key,
            max_length=max_length,
            play_beep=play_beep,
            trim=trim,
            recording_status_callback=recording_status_callback,
            recording_status_callback_method=recording_status_callback_method,
            transcribe=transcribe,
            transcribe_callback=transcribe_callback,
            **kwargs
        ))

    def redirect(self, url, method=None, **kwargs):
        """
        Add a <Redirect> element

        :param url: redirect url
        :param method: redirect method
        :param kwargs: additional attributes
        :return: <Response> element
        """
        return self.nest(Redirect(url, method=method, **kwargs))

    def reject(self, reason=None, **kwargs):
        """
        Add a <Reject> element

        :param reason: rejection reason
        :param kwargs: additional attributes
        :return: <Response> element
        """
        return self.nest(Reject(reason=reason, **kwargs))

    def say(self,
            body,
            loop=None,
            language=None,
            voice=None,
            **kwargs):
        """
        Add a <Say> element

        :param body: message body
        :param loop: times to loop
        :param language: language of message
        :param voice: voice to use
        :param kwargs: additional attributes
        :return: <Response> element
        """
        return self.nest(Say(
            body,
            loop=loop,
            language=language,
            voice=voice,
            **kwargs
        ))

    def sms(self,
            body,
            to=None,
            from_=None,
            method=None,
            action=None,
            status_callback=None,
            **kwargs):
        """
        Add a <Sms> element

        :param body: body of message
        :param to: to phone number
        :param from_: from phone number
        :param method: action URL method
        :param action: action URL
        :param status_callback: status callback URL
        :param kwargs: additional attributes
        :return: <Response> element
        """
        return self.nest(Sms(
            body,
            to=to,
            from_=from_,
            method=method,
            action=action,
            status_callback=status_callback,
            **kwargs
        ))


class Dial(TwiML):
    """
    <Dial> element
    """
    def __init__(self, number=None, **kwargs):
        """
        Create a new <Dial> element

        :param number: phone number to dial
        :param kwargs: additional attributes
        """
        super(Dial, self).__init__(**kwargs)
        if number:
            self.value = number

    def client(self,
               name,
               method=None,
               url=None,
               status_callback_event=None,
               status_callback_method=None,
               status_callback=None,
               **kwargs):
        """
        Add a new <Client> element

        :param name: name of client
        :param method: action URL method
        :param url: action URL
        :param status_callback_event: events to call status callback
        :param status_callback_method: status callback URL method
        :param status_callback: status callback URL
        :param kwargs: additional attributes
        :return: <Dial> element
        """
        return self.nest(Client(
            name,
            method=method,
            url=url,
            status_callback_event=status_callback_event,
            status_callback_method=status_callback_method,
            status_callback=status_callback,
            **kwargs
        ))

    def conference(self,
                   name,
                   muted=None,
                   start_conference_on_enter=None,
                   end_conference_on_exit=None,
                   max_participants=None,
                   beep=None,
                   record=None,
                   trim=None,
                   wait_method=None,
                   wait_url=None,
                   event_callback_url=None,
                   status_callback_event=None,
                   status_callback=None,
                   status_callback_method=None,
                   recording_status_callback=None,
                   recording_status_callback_method=None,
                   **kwargs):
        """
        Add a <Conference> element

        :param name: name of conference
        :param muted: join the conference muted
        :param start_conference_on_enter: start the conference on enter
        :param end_conference_on_exit: end the conference on exit
        :param max_participants: max number of people in conference
        :param beep: play beep when joining
        :param record: record the conference
        :param trim: trim the recording
        :param wait_method: wait URL method
        :param wait_url: wait URL to play
        :param event_callback_url: event callback URL
        :param status_callback_event: events to call status callback
        :param status_callback: status callback URL
        :param status_callback_method: status callback URL method
        :param recording_status_callback: recording status callback URL
        :param recording_status_callback_method: recording status callback URL method
        :param kwargs: additional attributes
        :return: <Dial> element
        """
        return self.nest(Conference(
            name,
            muted=muted,
            start_conference_on_enter=start_conference_on_enter,
            end_conference_on_exit=end_conference_on_exit,
            max_participants=max_participants,
            beep=beep,
            record=record,
            trim=trim,
            wait_method=wait_method,
            wait_url=wait_url,
            event_callback_url=event_callback_url,
            status_callback_event=status_callback_event,
            status_callback=status_callback,
            status_callback_method=status_callback_method,
            recording_status_callback=recording_status_callback,
            recording_status_callback_method=recording_status_callback_method,
            **kwargs
        ))

    def number(self,
               number,
               send_digits=None,
               url=None,
               method=None,
               status_callback_event=None,
               status_callback=None,
               status_callback_method=None,
               **kwargs):
        """
        Add a <Number> element

        :param number: phone number to dial
        :param send_digits: play DTMF tones when the call is answered
        :param url: TwiML URL
        :param method: TwiML URL method
        :param status_callback_event: events to call status callback
        :param status_callback: status callback URL
        :param status_callback_method: status callback URL method
        :param kwargs: additional attributes
        :return: <Dial> element
        """
        return self.nest(Number(
            number,
            send_digits=send_digits,
            url=url,
            method=method,
            status_callback_event=status_callback_event,
            status_callback=status_callback,
            status_callback_method=status_callback_method,
            **kwargs
        ))

    def queue(self,
              queue_name,
              url=None,
              method=None,
              reservation_sid=None,
              post_work_activity_sid=None,
              **kwargs):
        """
        Add a <Queue> element

        :param queue_name: queue name
        :param url: action URL
        :param method: action URL method
        :param reservation_sid: TaskRouter reservation SID
        :param post_work_activity_sid: TaskRouter activity SID
        :param kwargs: additional attributes
        :return: <Dial> element
        """
        return self.nest(Queue(
            queue_name,
            url=url,
            method=method,
            reservation_sid=reservation_sid,
            post_work_activity_sid=post_work_activity_sid,
            **kwargs
        ))

    def sim(self,
            sid,
            **kwargs):
        """
        Add a <Sim> element

        :param sid: sim sid
        :return: <Dial> element
        """
        return self.nest(Sim(
            sid,
            **kwargs
        ))

    def sip(self,
            uri,
            username=None,
            password=None,
            url=None,
            method=None,
            status_callback_event=None,
            status_callback=None,
            status_callback_method=None,
            **kwargs):
        """
        Add a <Sip> element

        :param uri: sip url
        :param username: sip username
        :param password: sip password
        :param url: action URL
        :param method: action URL method
        :param status_callback_event: events to call status callback
        :param status_callback: status callback URL
        :param status_callback_method: status callback URL method
        :param kwargs: additional attributes
        :return: <Dial> element
        """
        return self.nest(Sip(
            uri,
            username=username,
            password=password,
            url=url,
            method=method,
            status_callback_event=status_callback_event,
            status_callback=status_callback,
            status_callback_method=status_callback_method,
            **kwargs
        ))


class Client(TwiML):
    """
    <Client> element
    """
    def __init__(self, name, **kwargs):
        """
        Create a new <Client> element

        :param name: name of client
        :param kwargs: attributes
        """
        super(Client, self).__init__(**kwargs)
        self.value = name


class Conference(TwiML):
    """
    <Conference> element
    """
    def __init__(self, name, **kwargs):
        """
        Create a new <Conference> element

        :param name: name of conference
        :param kwargs: attributes
        """
        super(Conference, self).__init__(**kwargs)
        self.value = name


class Number(TwiML):
    """
    <Number> element
    """
    def __init__(self, number, **kwargs):
        """
        Create a new <Number> element

        :param number: phone number
        :param kwargs: attributes
        """
        super(Number, self).__init__(**kwargs)
        self.value = number


class Queue(TwiML):
    """
    <Queue> element
    """
    def __init__(self, queue_name, **kwargs):
        """
        Create a new <Queue> element

        :param queue_name: name of queue
        :param kwargs: attributes
        """
        super(Queue, self).__init__(**kwargs)
        self.value = queue_name


class Sim(TwiML):
    """
    <Sim> element
    """
    def __init__(self, sid, **kwargs):
        """
        Create a new <Sim> element

        :param sid: sim sid
        """
        super(Sim, self).__init__(**kwargs)
        self.value = sid


class Sip(TwiML):
    """
    <Sip> element
    """
    def __init__(self, uri, **kwargs):
        """
        Create a new <Sip> element

        :param uri: sip url
        :param kwargs: attributes
        """
        super(Sip, self).__init__(**kwargs)
        self.value = uri


class Echo(TwiML):
    """
    <Echo> element
    """
    def __init__(self, **kwargs):
        """
        Create a new <Echo> element
        """
        super(Echo, self).__init__(**kwargs)


class Enqueue(TwiML):
    """
    <Enqueue> element
    """
    def __init__(self, name, **kwargs):
        """
        Create a new <Enqueue> element

        :param name: queue name
        :param kwargs: attributes
        """
        super(Enqueue, self).__init__(**kwargs)
        self.value = name

    def task(self, attributes, **kwargs):
        """
        Add a <Task> element

        :param attributes: Attributes for a task
        :return: <Task> element
        """
        return self.nest(Task(attributes, **kwargs))


class Task(TwiML):
    """
    <Task> element
    """
    def __init__(self, attributes, **kwargs):
        """
        Create a new <Task> element

        :param attributes: Attributes for a task
        """
        super(Task, self).__init__(**kwargs)
        if isinstance(attributes, string_types):
            self.value = attributes
        else:
            self.value = json.dumps(attributes)


class Gather(TwiML):
    """
    <Gather> element
    """
    def __init__(self, **kwargs):
        """
        Create a new <Gather> element
        :param kwargs: attributes
        """
        super(Gather, self).__init__(**kwargs)

    def say(self,
            body,
            loop=None,
            language=None,
            voice=None,
            **kwargs):
        """
        Add a new <Say> element

        :param body: message body
        :param loop: times to loop
        :param language: message language
        :param voice: voice to use
        :param kwargs: additional attributes
        :return: <Gather> element
        """
        return self.nest(Say(
            body,
            loop=loop,
            language=language,
            voice=voice,
            **kwargs
        ))

    def play(self,
             url=None,
             loop=None,
             digits=None,
             **kwargs):
        """
        Add a new <Play> element

        :param url: media URL
        :param loop: times to loop
        :param digits: digits to simulate
        :param kwargs: additional attributes
        :return: <Gather> element
        """
        return self.nest(Play(
            url=url,
            loop=loop,
            digits=digits,
            **kwargs
        ))

    def pause(self, length=None):
        """
        Add a new <Pause> element

        :param length: time to pause
        :return: <Gather> element
        """
        return self.nest(Pause(length=length))


class Pause(TwiML):
    """
    <Pause> element
    """
    pass


class Play(TwiML):
    """
    <Play> element
    """
    def __init__(self, url=None, **kwargs):
        """
        Create a new <Play> element

        :param url: optional media URL
        :param kwargs: attributes
        """
        super(Play, self).__init__(**kwargs)
        self.value = url


class Say(TwiML):
    """
    <Say> element
    """
    def __init__(self, body, **kwargs):
        """
        Create a new <Say> element

        :param body: message body
        :param kwargs: attributes
        """
        super(Say, self).__init__(**kwargs)
        self.value = body


class Hangup(TwiML):
    """
    <Hangup> element
    """
    pass


class Leave(TwiML):
    """
    <Leave> element
    """
    pass


class Record(TwiML):
    """
    <Record> element
    """
    pass


class Redirect(TwiML):
    """
    <Redirect> element
    """
    def __init__(self, url, **kwargs):
        """
        Create a new <Redirect> element

        :param url: TwiML URL
        :param kwargs: attributes
        """
        super(Redirect, self).__init__(**kwargs)
        self.value = url


class Reject(TwiML):
    """
    <Reject> element
    """
    pass


class Sms(TwiML):
    """
    <Sms> element
    """
    def __init__(self, body, **kwargs):
        """
        Create a new <Sms> element

        :param body: message body
        :param kwargs: attributes
        """
        super(Sms, self).__init__(**kwargs)
        self.value = body



# coding=utf-8
"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page


class InstalledAddOnExtensionList(ListResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, installed_add_on_sid):
        """
        Initialize the InstalledAddOnExtensionList

        :param Version version: Version that contains the resource
        :param installed_add_on_sid: The installed_add_on_sid

        :returns: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionList
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionList
        """
        super(InstalledAddOnExtensionList, self).__init__(version)

        # Path Solution
        self._solution = {
            'installed_add_on_sid': installed_add_on_sid,
        }
        self._uri = '/InstalledAddOns/{installed_add_on_sid}/Extensions'.format(**self._solution)

    def stream(self, limit=None, page_size=None):
        """
        Streams InstalledAddOnExtensionInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionInstance]
        """
        limits = self._version.read_limits(limit, page_size)

        page = self.page(
            page_size=limits['page_size'],
        )

        return self._version.stream(page, limits['limit'], limits['page_limit'])

    def list(self, limit=None, page_size=None):
        """
        Lists InstalledAddOnExtensionInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionInstance]
        """
        return list(self.stream(
            limit=limit,
            page_size=page_size,
        ))

    def page(self, page_token=values.unset, page_number=values.unset,
             page_size=values.unset):
        """
        Retrieve a single page of InstalledAddOnExtensionInstance records from the API.
        Request is executed immediately

        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of InstalledAddOnExtensionInstance
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionPage
        """
        params = values.of({
            'PageToken': page_token,
            'Page': page_number,
            'PageSize': page_size,
        })

        response = self._version.page(
            'GET',
            self._uri,
            params=params,
        )

        return InstalledAddOnExtensionPage(self._version, response, self._solution)

    def get_page(self, target_url):
        """
        Retrieve a specific page of InstalledAddOnExtensionInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of InstalledAddOnExtensionInstance
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionPage
        """
        response = self._version.domain.twilio.request(
            'GET',
            target_url,
        )

        return InstalledAddOnExtensionPage(self._version, response, self._solution)

    def get(self, sid):
        """
        Constructs a InstalledAddOnExtensionContext

        :param sid: The unique Extension Sid

        :returns: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionContext
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionContext
        """
        return InstalledAddOnExtensionContext(
            self._version,
            installed_add_on_sid=self._solution['installed_add_on_sid'],
            sid=sid,
        )

    def __call__(self, sid):
        """
        Constructs a InstalledAddOnExtensionContext

        :param sid: The unique Extension Sid

        :returns: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionContext
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionContext
        """
        return InstalledAddOnExtensionContext(
            self._version,
            installed_add_on_sid=self._solution['installed_add_on_sid'],
            sid=sid,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Preview.Marketplace.InstalledAddOnExtensionList>'


class InstalledAddOnExtensionPage(Page):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, response, solution):
        """
        Initialize the InstalledAddOnExtensionPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API
        :param installed_add_on_sid: The installed_add_on_sid

        :returns: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionPage
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionPage
        """
        super(InstalledAddOnExtensionPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of InstalledAddOnExtensionInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionInstance
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionInstance
        """
        return InstalledAddOnExtensionInstance(
            self._version,
            payload,
            installed_add_on_sid=self._solution['installed_add_on_sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Preview.Marketplace.InstalledAddOnExtensionPage>'


class InstalledAddOnExtensionContext(InstanceContext):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, installed_add_on_sid, sid):
        """
        Initialize the InstalledAddOnExtensionContext

        :param Version version: Version that contains the resource
        :param installed_add_on_sid: The installed_add_on_sid
        :param sid: The unique Extension Sid

        :returns: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionContext
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionContext
        """
        super(InstalledAddOnExtensionContext, self).__init__(version)

        # Path Solution
        self._solution = {
            'installed_add_on_sid': installed_add_on_sid,
            'sid': sid,
        }
        self._uri = '/InstalledAddOns/{installed_add_on_sid}/Extensions/{sid}'.format(**self._solution)

    def fetch(self):
        """
        Fetch a InstalledAddOnExtensionInstance

        :returns: Fetched InstalledAddOnExtensionInstance
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionInstance
        """
        params = values.of({})

        payload = self._version.fetch(
            'GET',
            self._uri,
            params=params,
        )

        return InstalledAddOnExtensionInstance(
            self._version,
            payload,
            installed_add_on_sid=self._solution['installed_add_on_sid'],
            sid=self._solution['sid'],
        )

    def update(self, enabled):
        """
        Update the InstalledAddOnExtensionInstance

        :param bool enabled: A Boolean indicating if the Extension will be invoked

        :returns: Updated InstalledAddOnExtensionInstance
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionInstance
        """
        data = values.of({
            'Enabled': enabled,
        })

        payload = self._version.update(
            'POST',
            self._uri,
            data=data,
        )

        return InstalledAddOnExtensionInstance(
            self._version,
            payload,
            installed_add_on_sid=self._solution['installed_add_on_sid'],
            sid=self._solution['sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Preview.Marketplace.InstalledAddOnExtensionContext {}>'.format(context)


class InstalledAddOnExtensionInstance(InstanceResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, payload, installed_add_on_sid, sid=None):
        """
        Initialize the InstalledAddOnExtensionInstance

        :returns: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionInstance
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionInstance
        """
        super(InstalledAddOnExtensionInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'sid': payload['sid'],
            'installed_add_on_sid': payload['installed_add_on_sid'],
            'friendly_name': payload['friendly_name'],
            'product_name': payload['product_name'],
            'unique_name': payload['unique_name'],
            'enabled': payload['enabled'],
            'url': payload['url'],
        }

        # Context
        self._context = None
        self._solution = {
            'installed_add_on_sid': installed_add_on_sid,
            'sid': sid or self._properties['sid'],
        }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: InstalledAddOnExtensionContext for this InstalledAddOnExtensionInstance
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionContext
        """
        if self._context is None:
            self._context = InstalledAddOnExtensionContext(
                self._version,
                installed_add_on_sid=self._solution['installed_add_on_sid'],
                sid=self._solution['sid'],
            )
        return self._context

    @property
    def sid(self):
        """
        :returns: A string that uniquely identifies this Extension
        :rtype: unicode
        """
        return self._properties['sid']

    @property
    def installed_add_on_sid(self):
        """
        :returns: The installed_add_on_sid
        :rtype: unicode
        """
        return self._properties['installed_add_on_sid']

    @property
    def friendly_name(self):
        """
        :returns: A human-readable description of this Extension
        :rtype: unicode
        """
        return self._properties['friendly_name']

    @property
    def product_name(self):
        """
        :returns: A human-readable description of the Extension's Product
        :rtype: unicode
        """
        return self._properties['product_name']

    @property
    def unique_name(self):
        """
        :returns: The string that uniquely identifies this Extension
        :rtype: unicode
        """
        return self._properties['unique_name']

    @property
    def enabled(self):
        """
        :returns: A Boolean indicating if the Extension will be invoked
        :rtype: bool
        """
        return self._properties['enabled']

    @property
    def url(self):
        """
        :returns: The url
        :rtype: unicode
        """
        return self._properties['url']

    def fetch(self):
        """
        Fetch a InstalledAddOnExtensionInstance

        :returns: Fetched InstalledAddOnExtensionInstance
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionInstance
        """
        return self._proxy.fetch()

    def update(self, enabled):
        """
        Update the InstalledAddOnExtensionInstance

        :param bool enabled: A Boolean indicating if the Extension will be invoked

        :returns: Updated InstalledAddOnExtensionInstance
        :rtype: twilio.rest.preview.marketplace.installed_add_on.installed_add_on_extension.InstalledAddOnExtensionInstance
        """
        return self._proxy.update(
            enabled,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Preview.Marketplace.InstalledAddOnExtensionInstance {}>'.format(context)
